/*
 * In this flight all drone take off when start signal is sent. The leader
 * follows continuously a list of checkpoints. The other drones follow each
 * other (second drone follows leader, third drone follows second drone, ...).
 * When the switch signal is sent, the last drone becomes the leader and follow
 * the checkpoints in the reversed order. When stop signal is sent, each drone
 * lands one at the time from the last one to the leader.
*/

/*jslint nomen: true, indent: 2, maxerr: 3, maxlen: 80 */
/*global console, me*/
(function (console, me) {
  "use strict";

  var ALTITUDE_DIFF = 5,
    BASE_ALTITUDE = 25,
    EPSILON = 4,
    TARGETED_DISTANCE = 10,
    MIN_CMD_INTERVAL = 500,
    MIN_SPEED = 4,
    DEFAULT_SPEED = 5,
    MAX_SPEED = 6,
    R = 6371e3,
    SPEED_FACTOR = 0.2,
    TIMEOUT = 1000;

  function exitOnFail(ret, msg) {
    if (ret) {
      console.log(msg);
      me.exit(1);
    }
  }

  function toRad(angle) {
    return Math.PI * angle / 180;
  }

  function distance(lat1, lon1, lat2, lon2) {
    var la1 = toRad(lat1),
      la2 = toRad(lat2),
      lo1 = toRad(lon1),
      lo2 = toRad(lon2),
      haversine_phi = Math.pow(Math.sin((la2 - la1) / 2), 2),
      sin_lon = Math.sin((lo2 - lo1) / 2),
      h = haversine_phi + Math.cos(la1) * Math.cos(la2) * sin_lon * sin_lon;
    return 2 * R * Math.asin(Math.sqrt(h));
  }

  function leaderId(drone) {
    var id_list = drone.id_list;
    return (!drone.reverse) ? id_list[0] : id_list[id_list.length - 1];
  }

  function pointReached(drone, point, acceptance) {
    var current_position = drone.getCurrentPosition(),
      dist = distance(
        current_position.latitude,
        current_position.longitude,
        point.latitude,
        point.longitude
      );
    console.log("Distance to point", dist);
    return dist <= acceptance;
  }

  function goTo(drone, latitude, longitude, speed, timestamp) {
    if (timestamp - drone.last_cmd_timestamp >= MIN_CMD_INTERVAL) {
      drone.setTargetCoordinates(
        latitude,
        longitude,
        drone.flightAltitude,
        speed
      );
      drone.last_cmd_timestamp = timestamp;
    }
  }

  function communicate(drone, okState, state) {
    var infos = {id: drone.id, okState: okState, state: state};
    if (okState) {
      infos.timestamp = Date.now();
    }
    return drone.sendMsg(JSON.stringify(infos));
  }

  me.onStart = function () {
    me.direction_set = false;
    me.flightAltitude = BASE_ALTITUDE + me.id * ALTITUDE_DIFF;
    me.following = false;
    me.going_to_parachute = false;
    me.id_list = [me.id];
    me.timestamp_list = [0];
    me.landing = false;
    me.last_cmd_timestamp = 0;
    me.lost_a_drone = false;
    me.next_checkpoint = 0;
    me.started = false;
    me.stopped = false;
    me.reverse = false;
  };

  me.onUpdate = function (timestamp) {
    var checkpointIndex,
      distance2d,
      distanceDiff,
      distanceToTakeOffPoint,
      leader_id = leaderId(me),
      me_index = me.id_list.indexOf(me.id),
      neighbor_id = me.id_list[(!me.reverse) ? me_index - 1 : me_index + 1],
      neighbor_position,
      newSpeed,
      position = me.getCurrentPosition();

    if (leader_id !== me.id) {
      neighbor_position = me.getDroneDict()[neighbor_id];
    }

    if (!me.checkpoint_list) {
      if (leader_id !== me.id) {
        console.log(
          "timestamp difference",
          position.timestamp - neighbor_position.timestamp
        );
      }
      return communicate(me, true, "Ready");
    }

    if (me.landing && me.isLanding()) {
      return communicate(me, false, "Landing");
    }

    if (!me.started) {
      if (me.isReadyToFly()) {
        me.started = true;
      } else {
        return communicate(me, true, "Taking Off");
      }
    }

    if (!me.landing) {
      if (!me.isReadyToFly()) {
        return communicate(me, false, "Emergency");
      }

      JSON.parse(JSON.stringify(me.timestamp_list)).forEach(
        function (last_timestamp, index) {
          if (me.id_list[index] !== me.id
              && Math.abs(timestamp - last_timestamp) >= TIMEOUT) {
            console.log(
              "Lost drone",
              me.id_list[index],
              "timestamp difference",
              timestamp - last_timestamp
            );
            me.id_list.splice(index, 1);
            me.timestamp_list.splice(index, 1);
            me.lost_a_drone = true;
          }
        }
      );

      if (me.lost_a_drone) {
        me.lost_a_drone = false;
        return communicate(me, true, "Flying");
      }

      if (leader_id !== me.id) {
        distance2d = distance(
          position.latitude,
          position.longitude,
          neighbor_position.latitude,
          neighbor_position.longitude
        );

        if (!me.following) {
          distanceToTakeOffPoint = distance(
            neighbor_position.latitude,
            neighbor_position.longitude,
            me.landing_point_list[0].latitude,
            me.landing_point_list[0].longitude
          );
          console.log(
            "Distance from neighbor to takeoff point",
            distanceToTakeOffPoint
          );
          if (distanceToTakeOffPoint < 2 * TARGETED_DISTANCE) {
            return communicate(me, true, "Waiting");
          }
          me.following = true;
        }

        distanceDiff = distance2d - TARGETED_DISTANCE;
        newSpeed = Math.max(
          Math.min(distanceDiff * SPEED_FACTOR + DEFAULT_SPEED, MAX_SPEED),
          MIN_SPEED
        );

        goTo(
          me,
          neighbor_position.latitude,
          neighbor_position.longitude,
          newSpeed,
          timestamp
        );

        console.log("distance to leader", distance2d);
        return communicate(me, true, "Flying");
      }

      checkpointIndex = (!me.reverse) ? me.next_checkpoint
        : me.checkpoint_list.length - me.next_checkpoint - 1;
      if (!me.direction_set) {
        console.log("Going to Checkpoint", checkpointIndex);
        me.next_point = {
          "latitude": me.checkpoint_list[checkpointIndex].latitude,
          "longitude": me.checkpoint_list[checkpointIndex].longitude
        };
        me.direction_set = true;
      }

      if (
        pointReached(me, me.checkpoint_list[checkpointIndex], EPSILON)
      ) {
        console.log("Reached Checkpoint", checkpointIndex);
        me.next_checkpoint += 1;
        me.next_checkpoint %= me.checkpoint_list.length;
        me.direction_set = false;
        return me.sendMsg(JSON.stringify({
          id: me.id,
          okState: true,
          next_checkpoint: me.next_checkpoint,
          state: "Flying",
          timestamp: Date.now()
        }));
      }
    }

    if (me.going_to_parachute &&
        pointReached(me, me.landing_point_list[me.id], EPSILON)) {
      me.going_to_parachute = false;
      return exitOnFail(me.land(), "Failed to land");
    }

    if (!me.landing || me.going_to_parachute) {
      goTo(
        me,
        me.next_point.latitude,
        me.next_point.longitude,
        DEFAULT_SPEED,
        Date.now()
      );
    }

    if (me.landing) {
      return communicate(me, false, "Landed");
    } else {
      return communicate(me, true, "Flying");
    }
  };

  me.onGetMsg = function (msg) {
    if (msg.hasOwnProperty("checkpoint_list")) {
      me.checkpoint_list = msg.checkpoint_list;
      me.landing_point_list = msg.landing_point_list;
      console.log("running");
      return me.takeOff();
    }

    var msgDict = JSON.parse(msg);

    if (msgDict.hasOwnProperty("status")) {
      switch (msgDict.status) {

      case "stopped":
        me.stopped = true;
        if (me.id ===
            (!me.reverse ? me.id_list[me.id_list.length - 1] : me.id_list[0])) {
          me.landing = true;
          me.going_to_parachute = true;
          me.next_point = {
            "latitude": me.landing_point_list[me.id].latitude,
            "longitude": me.landing_point_list[me.id].longitude
          };
          console.log("Going to rally");
        }
        break;

      case "switch":
        me.reverse = !me.reverse;
        me.direction_set = false;
        break;

      default:
        console.log("Unknown status:", msgDict.status);
      }
      return;
    }

    if (msgDict.hasOwnProperty("state")) {
      switch (msgDict.state) {

      case "Ready":
        if (!me.id_list.includes(msgDict.id)) {
          me.id_list[me.id_list.length] = msgDict.id;
          me.id_list.sort();
        }
        break;

      case "Landing":
        if (msgDict.id !== me.id && me.id_list.includes(msgDict.id)) {
          me.id_list.splice(me.id_list.indexOf(msgDict.id), 1);
          me.timestamp_list.splice(me.id_list.indexOf(msgDict.id), 1);
          if (me.stopped && me.id ===
              (!me.reverse ? me.id_list[me.id_list.length - 1] : me.id_list[0])
              ) {
            me.landing = true;
            me.going_to_parachute = true;
            me.next_point = {
              "latitude": me.landing_point_list[me.id].latitude,
              "longitude": me.landing_point_list[me.id].longitude
            };
          }
        }
        break;

      default:
        break;
      }
    }

    if (me.id_list.includes(msgDict.id)
        && msgDict.hasOwnProperty("timestamp")) {
      me.timestamp_list[me.id_list.indexOf(msgDict.id)] = msgDict.timestamp;
    }

    if (msgDict.hasOwnProperty("next_checkpoint")) {
      me.next_checkpoint = msgDict.next_checkpoint;
    }
  };
}(console, me));
